#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_snmp_mplsl3vpn.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2020 by Volta Networks
#

"""
test_bgp_snmp_mplsl3vpn.py: Test mplsL3Vpn MIB [RFC4382].
"""

import os
import sys
from time import sleep
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.snmptest import SnmpTester
from lib import topotest

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.bgpd, pytest.mark.isisd, pytest.mark.snmp]


def build_topo(tgen):
    "Build function"

    # This function only purpose is to define allocation and relationship
    # between routers, switches and hosts.
    #
    #
    # Create routers
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")
    tgen.add_router("r4")
    tgen.add_router("ce1")
    tgen.add_router("ce2")
    tgen.add_router("ce3")
    tgen.add_router("ce4")

    # r1-r2
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # r1-r3
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])

    # r1-r4
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r4"])

    # r1-ce1
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["ce1"])

    # r1-ce3
    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["ce3"])

    # r1-ce4
    switch = tgen.add_switch("s6")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["ce4"])

    # r1-dangling
    switch = tgen.add_switch("s7")
    switch.add_link(tgen.gears["r1"])

    # r2-r3
    switch = tgen.add_switch("s8")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    # r3-r4
    switch = tgen.add_switch("s9")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r4"])

    # r4-ce2
    switch = tgen.add_switch("s10")
    switch.add_link(tgen.gears["r4"])
    switch.add_link(tgen.gears["ce2"])


def setup_module(mod):
    "Sets up the pytest environment"

    # skip tests is SNMP not installed
    snmpd = os.system("which snmpd")
    if snmpd:
        error_msg = "SNMP not installed - skipping"
        pytest.skip(error_msg)

    # This function initiates the topology build with Topogen...
    tgen = Topogen(build_topo, mod.__name__)
    # ... and here it calls Mininet initialization functions.
    tgen.start_topology()

    r1 = tgen.gears["r1"]
    r2 = tgen.gears["r2"]
    r3 = tgen.gears["r3"]
    r4 = tgen.gears["r4"]

    # setup VRF-a in r1
    r1.run("ip link add VRF-a type vrf table 1001")
    r1.run("ip link set up dev VRF-a")
    r1.run("ip link add VRF-b type vrf table 1002")
    r1.run("ip link set up dev VRF-b")
    r4.run("ip link add VRF-a type vrf table 1001")
    r4.run("ip link set up dev VRF-a")

    # enslave vrf interfaces
    r1.run("ip link set r1-eth3 master VRF-a")
    r1.run("ip link set r1-eth4 master VRF-a")
    r1.run("ip link set r1-eth5 master VRF-b")
    r4.run("ip link set r4-eth1 master VRF-a")

    r1.run("sysctl -w net.ipv4.ip_forward=1")
    r2.run("sysctl -w net.ipv4.ip_forward=1")
    r3.run("sysctl -w net.ipv4.ip_forward=1")
    r4.run("sysctl -w net.ipv4.ip_forward=1")
    r1.run("sysctl -w net.mpls.conf.r1-eth0.input=1")
    r1.run("sysctl -w net.mpls.conf.r1-eth1.input=1")
    r1.run("sysctl -w net.mpls.conf.r1-eth2.input=1")

    router_list = tgen.routers()

    # For all registered routers, load the zebra configuration file
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP,
            os.path.join(CWD, "{}/bgpd.conf".format(rname)),
            "-M snmp",
        )
        router.load_config(
            TopoRouter.RD_SNMP,
            os.path.join(CWD, "{}/snmpd.conf".format(rname)),
            "-Le -Ivacm_conf,usmConf,iquery -V -DAgentX,trap",
        )

    # After loading the configurations, this function loads configured daemons.
    tgen.start_router()


def teardown_module(mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


# SNMP utilities - maybe move to lib
def snmp_uint32_to_oid(val):
    oid1 = int(val / 16777216) % 256
    oid2 = int(val / 65536) % 256
    oid3 = int(val / 256) % 256
    oid4 = int(val) % 256
    return "%(oid1)s.%(oid2)s.%(oid3)s.%(oid4)s" % locals()


def snmp_oid_to_uint32(oid):
    values = oid.split(".")
    return (
        (int(values[0]) * 16777216)
        + (int(values[1]) * 65536)
        + (int(values[2]) * 256)
        + int(values[3])
    )


def snmp_str_to_oid(str):
    out_oid = ""
    for char in str:
        out_oid += "{}.".format(ord(char))
    return out_oid.rstrip(".")


def snmp_oid_to_str(oid):
    out_str = ""
    oids = oid.split(".")
    for char in oids:
        out_str += "{}".format(chr(int(char)))
    return out_str


def snmp_rte_oid(vrf, dtype, dest, plen, policy, ntype, nhop=0):
    oid_1 = snmp_str_to_oid(vrf)
    oid_2 = dtype
    oid_3 = dest
    oid_4 = plen
    oid_5 = "0.{}".format(policy)
    oid_6 = ntype
    if ntype == 0:
        oid_7 = ""
    else:
        oid_7 = ".{}".format(nhop)

    return "{}.{}.{}.{}.{}.{}{}".format(oid_1, oid_2, oid_3, oid_4, oid_5, oid_6, oid_7)


def test_pe1_converge_evpn():
    "Wait for protocol convergence"
    tgen = get_topogen()

    r1 = tgen.gears["r1"]

    def _convergence():
        r1 = tgen.gears["r1"]
        r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

        return r1_snmp.test_oid("bgpVersion", "10")

    _, result = topotest.run_and_expect(_convergence, True, count=20, wait=1)
    assertmsg = "BGP SNMP does not seem to be running"
    assert result, assertmsg

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")
    count = 0
    passed = False
    while count < 125:
        if r1_snmp.test_oid_walk("bgpPeerLocalAddr.10.4.4.4", ["10.1.1.1"]):
            passed = True
            break
        count += 1
        sleep(1)
    # tgen.mininet_cli()
    assertmsg = "BGP Peer 10.4.4.4 did not connect"
    assert passed, assertmsg


interfaces_up_test = {
    "mplsL3VpnConfiguredVrfs": "2",
    "mplsL3VpnActiveVrfs": "2",
    "mplsL3VpnConnectedInterfaces": "3",
    "mplsL3VpnNotificationEnable": "true(1)",
    "mplsL3VpnVrfConfMaxPossRts": "0",
    "mplsL3VpnVrfConfRteMxThrshTime": "0 seconds",
    "mplsL3VpnIlllblRcvThrsh": "0",
}

interfaces_down_test = {
    "mplsL3VpnConfiguredVrfs": "2",
    "mplsL3VpnActiveVrfs": "1",
    "mplsL3VpnConnectedInterfaces": "3",
    "mplsL3VpnNotificationEnable": "true(1)",
    "mplsL3VpnVrfConfMaxPossRts": "0",
    "mplsL3VpnVrfConfRteMxThrshTime": "0 seconds",
    "mplsL3VpnIlllblRcvThrsh": "0",
}


def test_r1_mplsvpn_scalars():
    "check scalar values"
    tgen = get_topogen()
    r1 = tgen.gears["r1"]
    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    for item in interfaces_up_test.keys():
        assertmsg = "{} should be {}: value {}".format(
            item, interfaces_up_test[item], r1_snmp.get_next(item)
        )
        assert r1_snmp.test_oid(item, interfaces_up_test[item]), assertmsg


def test_r1_mplsvpn_scalars_interface():
    "check scalar interface changing values"
    tgen = get_topogen()
    r1 = tgen.gears["r1"]
    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    r1.vtysh_cmd("conf t\ninterface r1-eth3\nshutdown")
    r1.vtysh_cmd("conf t\ninterface r1-eth4\nshutdown")

    for item in interfaces_up_test.keys():
        assertmsg = "{} should be {}: value {}".format(
            item, interfaces_down_test[item], r1_snmp.get_next(item)
        )
        assert r1_snmp.test_oid(item, interfaces_down_test[item]), assertmsg

    r1.vtysh_cmd("conf t\ninterface r1-eth3\nno shutdown")
    r1.vtysh_cmd("conf t\ninterface r1-eth4\nno shutdown")

    for item in interfaces_up_test.keys():
        assertmsg = "{} should be {}: value {}".format(
            item, interfaces_up_test[item], r1_snmp.get_next(item)
        )
        assert r1_snmp.test_oid(item, interfaces_up_test[item]), assertmsg


def router_interface_get_ifindex(router, interface):
    ifindex = 0
    r_int_output = router.vtysh_cmd(
        "show interface {}-{}".format(router.name, interface)
    )
    int_lines = r_int_output.splitlines()
    for line in int_lines:
        line_items = line.lstrip().split(" ")
        if "index" in line_items[0]:
            ifindex = line_items[1]
    return ifindex


def generate_vrf_ifindex_oid(vrf, ifindex):
    intoid = snmp_uint32_to_oid(int(ifindex))
    vrfoid = snmp_str_to_oid(vrf)
    oid = "{}.{}".format(vrfoid, intoid)

    return oid


def generate_vrf_index_type_oid(vrf, index, type):
    vrfoid = snmp_str_to_oid(vrf)
    intoid = snmp_uint32_to_oid(int(index))
    oid = "{}.{}.{}".format(vrfoid, intoid, type)

    return oid


iftable_up_test = {
    "mplsL3VpnIfVpnClassification": ["enterprise(2)", "enterprise(2)", "enterprise(2)"],
    "mplsL3VpnIfConfStorageType": ["volatile(2)", "volatile(2)", "volatile(2)"],
    "mplsL3VpnIfConfRowStatus": ["active(1)", "active(1)", "active(1)"],
}


def get_timetick_val(time):
    return int(time.split(" ")[0].lstrip("(").rstrip(")"))


def test_r1_mplsvpn_IfTable():
    "mplsL3VpnIf table values"

    tgen = get_topogen()
    r1 = tgen.gears["r1"]

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    # tgen.mininet_cli()
    eth3_ifindex = router_interface_get_ifindex(r1, "eth3")
    eth4_ifindex = router_interface_get_ifindex(r1, "eth4")
    eth5_ifindex = router_interface_get_ifindex(r1, "eth5")

    # get ifindex and make sure the oid is correct

    oids = []
    # generate oid
    oids.append(generate_vrf_ifindex_oid("VRF-a", eth3_ifindex))
    oids.append(generate_vrf_ifindex_oid("VRF-a", eth4_ifindex))
    oids.append(generate_vrf_ifindex_oid("VRF-b", eth5_ifindex))

    for item in iftable_up_test.keys():
        assertmsg = "{} should be {} oids {} full dict {}:".format(
            item, iftable_up_test[item], oids, r1_snmp.walk(item)
        )
        assert r1_snmp.test_oid_walk(item, iftable_up_test[item], oids), assertmsg

    # an inactive vrf should not affect these values
    r1.cmd("ip link set r1-eth5 down")

    for item in iftable_up_test.keys():
        assertmsg = "{} should be {} oids {} full dict {}:".format(
            item, iftable_up_test[item], oids, r1_snmp.walk(item)
        )
        assert r1_snmp.test_oid_walk(item, iftable_up_test[item], oids), assertmsg

    r1.cmd("ip link set r1-eth5 up")


vrftable_test = {
    "mplsL3VpnVrfDescription": ["VRF-a", "VRF-b"],
    "mplsL3VpnVrfRD": ['"10:1"', '"10:2"'],
    "mplsL3VpnVrfOperStatus": ["up(1)", "up(1)"],
    "mplsL3VpnVrfActiveInterfaces": ["2", "1"],
    "mplsL3VpnVrfAssociatedInterfaces": ["2", "1"],
    "mplsL3VpnVrfConfMidRteThresh": ["0", "0"],
    "mplsL3VpnVrfConfHighRteThresh": ["0", "0"],
    "mplsL3VpnVrfConfMaxRoutes": ["0", "0"],
    "mplsL3VpnVrfConfRowStatus": ["active(1)", "active(1)"],
    "mplsL3VpnVrfConfAdminStatus": ["up(1)", "up(1)"],
    "mplsL3VpnVrfConfStorageType": ["volatile(2)", "volatile(2)"],
}


def test_r1_mplsvpn_VrfTable():
    tgen = get_topogen()

    r1 = tgen.gears["r1"]

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    # tgen.mininet_cli()

    oids = []

    oids.append(snmp_str_to_oid("VRF-a"))
    oids.append(snmp_str_to_oid("VRF-b"))

    # check items
    for item in vrftable_test.keys():
        assertmsg = "{} should be {} oids {} full dict {}:".format(
            item, vrftable_test[item], oids, r1_snmp.walk(item)
        )
        assert r1_snmp.test_oid_walk(item, vrftable_test[item], oids), assertmsg

    # check timetick set and stable
    ts_a = r1_snmp.get("mplsL3VpnVrfCreationTime.{}".format(snmp_str_to_oid("VRF-a")))
    ts_b = r1_snmp.get("mplsL3VpnVrfCreationTime.{}".format(snmp_str_to_oid("VRF-b")))
    ts_val_a1 = get_timetick_val(ts_a)
    ts_val_b1 = get_timetick_val(ts_b)
    ts_a = r1_snmp.get("mplsL3VpnVrfCreationTime.{}".format(snmp_str_to_oid("VRF-a")))
    ts_b = r1_snmp.get("mplsL3VpnVrfCreationTime.{}".format(snmp_str_to_oid("VRF-b")))
    ts_val_a2 = get_timetick_val(ts_a)
    ts_val_b2 = get_timetick_val(ts_b)

    assertmsg = "timestamp values for VRF-a do not match {} {}".format(
        ts_val_a1, ts_val_a2
    )
    assert ts_val_a1 == ts_val_a2, assertmsg
    assertmsg = "timestamp values for VRF-b do not match {} {}".format(
        ts_val_b1, ts_val_b2
    )
    assert ts_val_b1 == ts_val_b2, assertmsg

    # take Last changed time, fiddle with active interfaces, ensure
    # time changes and active interfaces change
    ts_last = r1_snmp.get(
        "mplsL3VpnVrfConfLastChanged.{}".format(snmp_str_to_oid("VRF-a"))
    )
    ts_val_last_1 = get_timetick_val(ts_last)
    r1.vtysh_cmd("conf t\ninterface r1-eth3\nshutdown")
    active_int = r1_snmp.get(
        "mplsL3VpnVrfActiveInterfaces.{}".format(snmp_str_to_oid("VRF-a"))
    )
    assertmsg = "mplsL3VpnVrfActiveInterfaces incorrect should be 1 value {}".format(
        active_int
    )
    assert active_int == "1", assertmsg

    ts_last = r1_snmp.get(
        "mplsL3VpnVrfConfLastChanged.{}".format(snmp_str_to_oid("VRF-a"))
    )
    ts_val_last_2 = get_timetick_val(ts_last)
    assertmsg = "mplsL3VpnVrfConfLastChanged does not update on interface change"
    assert ts_val_last_2 > ts_val_last_1, assertmsg
    r1.vtysh_cmd("conf t\ninterface r1-eth3\nno shutdown")

    # take Last changed time, fiddle with associated interfaces, ensure
    # time changes and active interfaces change
    ts_last = r1_snmp.get(
        "mplsL3VpnVrfConfLastChanged.{}".format(snmp_str_to_oid("VRF-a"))
    )
    ts_val_last_1 = get_timetick_val(ts_last)
    r1.cmd("ip link set r1-eth6 master VRF-a")
    r1.cmd("ip link set r1-eth6 up")

    associated_int = r1_snmp.get(
        "mplsL3VpnVrfAssociatedInterfaces.{}".format(snmp_str_to_oid("VRF-a"))
    )
    assertmsg = (
        "mplsL3VpnVrfAssociatedInterfaces incorrect should be 3 value {}".format(
            associated_int
        )
    )

    assert associated_int == "3", assertmsg
    ts_last = r1_snmp.get(
        "mplsL3VpnVrfConfLastChanged.{}".format(snmp_str_to_oid("VRF-a"))
    )
    ts_val_last_2 = get_timetick_val(ts_last)
    assertmsg = "mplsL3VpnVrfConfLastChanged does not update on interface change"
    assert ts_val_last_2 > ts_val_last_1, assertmsg
    r1.cmd("ip link del r1-eth6 master VRF-a")
    r1.cmd("ip link set r1-eth6 down")


rt_table_test = {
    "mplsL3VpnVrfRT": ['"1:1"', '"1:2"'],
    "mplsL3VpnVrfRTDescr": ["RT both for VRF VRF-a", "RT both for VRF VRF-b"],
    "mplsL3VpnVrfRTRowStatus": ["active(1)", "active(1)"],
    "mplsL3VpnVrfRTStorageType": ["volatile(2)", "volatile(2)"],
}


def test_r1_mplsvpn_VrfRT_table():
    tgen = get_topogen()

    r1 = tgen.gears["r1"]

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    oids = []
    oids.append(generate_vrf_index_type_oid("VRF-a", 1, 3))
    oids.append(generate_vrf_index_type_oid("VRF-b", 1, 3))

    # check items
    for item in rt_table_test.keys():
        print(item)
        assertmsg = "{} should be {} oids {} full dict {}:".format(
            item, rt_table_test[item], oids, r1_snmp.walk(item)
        )
        assert r1_snmp.test_oid_walk(item, rt_table_test[item], oids), assertmsg


def test_r1_mplsvpn_perf_table():
    tgen = get_topogen()

    r1 = tgen.gears["r1"]

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    # tgen.mininet_cli()
    oid_a = snmp_str_to_oid("VRF-a")
    oid_b = snmp_str_to_oid("VRF-b")

    # poll for 10 seconds for routes to appear
    count = 0
    passed = False
    while count < 60:
        if r1_snmp.test_oid_walk(
            "mplsL3VpnVrfPerfCurrNumRoutes.{}".format(oid_a), ["7"]
        ):
            passed = True
            break
        count += 1
        sleep(1)
    # tgen.mininet_cli()
    assertmsg = "mplsL3VpnVrfPerfCurrNumRoutes shouold be 7 got {}".format(
        r1_snmp.get("mplsL3VpnVrfPerfCurrNumRoutes.{}".format(oid_a))
    )
    assert passed, assertmsg
    curr_a = int(r1_snmp.get("mplsL3VpnVrfPerfCurrNumRoutes.{}".format(oid_a)))
    del_a = int(r1_snmp.get("mplsL3VpnVrfPerfRoutesDeleted.{}".format(oid_a)))
    add_a = int(r1_snmp.get("mplsL3VpnVrfPerfRoutesAdded.{}".format(oid_a)))

    assertmsg = "FAIL curr{} does not equal added{} - deleted {}".format(
        curr_a, add_a, del_a
    )
    assert curr_a == (add_a - del_a), assertmsg
    curr_b = int(r1_snmp.get("mplsL3VpnVrfPerfCurrNumRoutes.{}".format(oid_b)))
    del_b = int(r1_snmp.get("mplsL3VpnVrfPerfRoutesDeleted.{}".format(oid_b)))
    add_b = int(r1_snmp.get("mplsL3VpnVrfPerfRoutesAdded.{}".format(oid_b)))
    assertmsg = "FAIL curr{} does not equal added{} - deleted {}".format(
        curr_b, add_b, del_b
    )
    assert curr_b == (add_b - del_b), assertmsg


rte_table_test = {
    "mplsL3VpnVrfRteInetCidrDestType": [
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
    ],
    "mplsL3VpnVrfRteInetCidrDest": [
        "0A 05 05 05",
        "0A 07 07 07",
        "C0 A8 22 00",
        "C0 A8 64 00",
        "C0 A8 64 00",
        "C0 A8 C8 00",
        "C0 A8 C8 00",
    ],
    "mplsL3VpnVrfRteInetCidrPfxLen": ["32", "32", "24", "24", "24", "24", "24"],
    "mplsL3VpnVrfRteInetCidrNHopType": [
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "ipv4(1)",
        "unknown(0)",
        "ipv4(1)",
        "unknown(0)",
    ],
    "mplsL3VpnVrfRteInetCidrNextHop": [
        "C0 A8 64 0A",
        "C0 A8 C8 0A",
        "0A 04 04 04",
        "C0 A8 64 0A",
        '""',
        "C0 A8 C8 0A",
        '""',
    ],
    "mplsL3VpnVrfRteInetCidrType": [
        "local(3)",
        "local(3)",
        "remote(4)",
        "local(3)",
        "other(1)",
        "local(3)",
        "other(1)",
    ],
    "mplsL3VpnVrfRteInetCidrProto": [
        "bgp(14)",
        "bgp(14)",
        "bgp(14)",
        "bgp(14)",
        "local(2)",
        "bgp(14)",
        "local(2)",
    ],
    "mplsL3VpnVrfRteInetCidrNextHopAS": [
        "65001",
        "65001",
        "0",
        "65001",
        "0",
        "65001",
        "0",
    ],
    "mplsL3VpnVrfRteInetCidrMetric1": ["0", "0", "20", "0", "0", "0", "0"],
    "mplsL3VpnVrfRteInetCidrMetric2": ["-1", "-1", "-1", "-1", "-1", "-1", "-1"],
    "mplsL3VpnVrfRteInetCidrMetric3": ["-1", "-1", "-1", "-1", "-1", "-1", "-1"],
    "mplsL3VpnVrfRteInetCidrMetric4": ["-1", "-1", "-1", "-1", "-1", "-1", "-1"],
    "mplsL3VpnVrfRteInetCidrMetric5": ["-1", "-1", "-1", "-1", "-1", "-1", "-1"],
    "mplsL3VpnVrfRteXCPointer": ["00", "00", "00", "00", "00", "00", "00"],
    "mplsL3VpnVrfRteInetCidrStatus": [
        "active(1)",
        "active(1)",
        "active(1)",
        "active(1)",
        "active(1)",
        "active(1)",
        "active(1)",
    ],
}


def test_r1_mplsvpn_rte_table():
    tgen = get_topogen()

    r1 = tgen.gears["r1"]

    r1_snmp = SnmpTester(r1, "10.1.1.1", "public", "2c")

    # tgen.mininet_cli()
    oid_1 = snmp_rte_oid("VRF-a", 1, "10.5.5.5", 32, 0, 1, "192.168.100.10")
    oid_2 = snmp_rte_oid("VRF-a", 1, "10.7.7.7", 32, 0, 1, "192.168.200.10")
    oid_3 = snmp_rte_oid("VRF-a", 1, "192.168.34.0", 24, 0, 1, "10.4.4.4")
    oid_4 = snmp_rte_oid("VRF-a", 1, "192.168.100.0", 24, 1, 1, "192.168.100.10")
    oid_4_a = snmp_rte_oid("VRF-a", 1, "192.168.100.0", 24, 0, 1, "192.168.100.10")
    oid_5 = snmp_rte_oid("VRF-a", 1, "192.168.100.0", 24, 0, 0)
    oid_5_a = snmp_rte_oid("VRF-a", 1, "192.168.100.0", 24, 1, 0)
    oid_6 = snmp_rte_oid("VRF-a", 1, "192.168.200.0", 24, 1, 1, "192.168.200.10")
    oid_6_a = snmp_rte_oid("VRF-a", 1, "192.168.200.0", 24, 0, 1, "192.168.200.10")
    oid_7 = snmp_rte_oid("VRF-a", 1, "192.168.200.0", 24, 0, 0)
    oid_7_a = snmp_rte_oid("VRF-a", 1, "192.168.200.0", 24, 1, 0)

    oid_lists = [
        [oid_1, oid_2, oid_3, oid_4, oid_5, oid_6, oid_7],
        [oid_1, oid_2, oid_3, oid_4_a, oid_5_a, oid_6, oid_7],
        [oid_1, oid_2, oid_3, oid_4, oid_5, oid_6_a, oid_7_a],
        [oid_1, oid_2, oid_3, oid_4_a, oid_5_a, oid_6_a, oid_7_a],
        [oid_1, oid_2, oid_3, oid_4, oid_5, oid_6, oid_7],
        [oid_1, oid_2, oid_3, oid_4_a, oid_5_a, oid_6, oid_7],
        [oid_1, oid_2, oid_3, oid_4, oid_5, oid_6_a, oid_7_a],
        [oid_1, oid_2, oid_3, oid_4_a, oid_5_a, oid_6_a, oid_7_a],
    ]

    # check items

    passed = False
    for oid_list in oid_lists:
        passed = True
        for item in rte_table_test.keys():
            print(item)
            assertmsg = "{} should be {} oids {} full dict {}:".format(
                item, rte_table_test[item], oid_list, r1_snmp.walk(item)
            )
            if not r1_snmp.test_oid_walk(item, rte_table_test[item], oid_list):
                passed = False
                break
            print(
                "{} should be {} oids {} full dict {}:".format(
                    item, rte_table_test[item], oid_list, r1_snmp.walk(item)
                )
            )
        if passed:
            break
    # generate ifindex row grabbing ifindices from vtysh
    if passed:
        ifindex_row = [
            router_interface_get_ifindex(r1, "eth3"),
            router_interface_get_ifindex(r1, "eth4"),
            router_interface_get_ifindex(r1, "eth2"),
            router_interface_get_ifindex(r1, "eth3"),
            "0",
            router_interface_get_ifindex(r1, "eth4"),
            "0",
        ]
        if not r1_snmp.test_oid_walk(
            "mplsL3VpnVrfRteInetCidrIfIndex", ifindex_row, oid_list
        ):
            passed = False

    print("passed {}".format(passed))
    assert passed, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# common_check.py
#
# Copyright 2024 6WIND S.A.

#
import json
from lib import topotest


def ip_check_path_selection(
    router, ipaddr_str, expected, vrf_name=None, check_fib=False
):
    if vrf_name:
        cmdstr = f"show ip route vrf {vrf_name} {ipaddr_str} json"
    else:
        cmdstr = f"show ip route {ipaddr_str} json"
    try:
        output = json.loads(router.vtysh_cmd(cmdstr))
    except:
        output = {}

    ret = topotest.json_cmp(output, expected)
    if ret is None:
        num_nh_expected = len(expected[ipaddr_str][0]["nexthops"])
        num_nh_observed = len(output[ipaddr_str][0]["nexthops"])
        if num_nh_expected == num_nh_observed:
            if check_fib:
                # special case: when fib flag is unset,
                # an extra test should be done to check that the flag is really unset
                for nh_output, nh_expected in zip(
                    output[ipaddr_str][0]["nexthops"],
                    expected[ipaddr_str][0]["nexthops"],
                ):
                    if (
                        "fib" in nh_output.keys()
                        and nh_output["fib"]
                        and ("fib" not in nh_expected.keys() or not nh_expected["fib"])
                    ):
                        return "{}, prefix {} nexthop {} has the fib flag set, whereas it is not expected".format(
                            router.name, ipaddr_str, nh_output["ip"]
                        )
            return ret
        return "{}, prefix {} does not have the correct number of nexthops : observed {}, expected {}".format(
            router.name, ipaddr_str, num_nh_observed, num_nh_expected
        )
    return ret


def iproute2_check_path_selection(router, ipaddr_str, expected, vrf_name=None):
    if not topotest.iproute2_is_json_capable():
        return None

    if vrf_name:
        cmdstr = f"ip -json route show vrf {vrf_name} {ipaddr_str}"
    else:
        cmdstr = f"ip -json route show {ipaddr_str}"
    try:
        output = json.loads(router.cmd(cmdstr))
    except:
        output = []

    return topotest.json_cmp(output, expected)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2019 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test Multicast basic functionality:

Topology:

                 _______r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5
                |             |
                |_____________|
                        r4

Test steps
- Create topology (setup module)
- Bring up topology

TC_1 : Verify upstream interfaces(IIF) and join state are updated properly
       after adding and deleting the static RP
TC_2 : Verify IIF and OIL in "show ip pim state" updated properly after
       adding and deleting the static RP
TC_3: (*, G) Mroute entry are cleared when static RP gets deleted
TC_4: Verify (*,G) prune is send towards the RP after deleting the static RP
TC_5: Verify OIF entry for RP is cleared when RP becomes unreachable
TC_6: Verify IIF and OIL in "show ip pim state" updated properly when RP
      becomes unreachable
TC_7 : Verify upstream interfaces(IIF) and join state are updated properly
       after adding and deleting the static RP
TC_8: Verify (*,G) prune is send towards the RP when RP becomes unreachable
TC_9 : Verify RP configured after IGMP join received, PIM join towards RP is
       sent immediately
TC_10 : Verify RP becomes reachable after IGMP join received, PIM join
        towards RP is sent immediately
TC_11 : Verify PIM join send towards the higher preferred RP
TC_12 : Verify PIM prune send towards the lower preferred RP
TC_13 : Verify RPF interface is updated in mroute (kernel) when higher
        preferred overlapping RP configured
TC_14 : Verify IIF and OIL in "show ip pim state" updated properly when higher
        preferred overlapping RP configured
TC_15 : Verify upstream interfaces(IIF) and join state are updated when higher
        preferred overlapping RP is configured
TC_16 : Verify join is send to lower preferred RP, when higher preferred RP
        gets deleted
TC_17 : Verify prune is send to higher preferred RP when higher preferred RP
        gets deleted
TC_18 : Verify RPF interface updated in mroute when higher preferred RP gets
        deleted
TC_19 : Verify IIF and OIL in "show ip pim state" updated when higher
        preferred overlapping RP is deleted
TC_20 : Verify PIM upstream IIF updated when higher preferred overlapping RP
        deleted
TC_21_1 : Verify OIF and RFP for (*,G) and (S,G) when static RP configure in
          LHR router
TC_21_2 : Verify OIF and RFP for (*,G) and (S,G) when static RP configure in
          LHR router
TC_22_1 : Verify OIF and RPF for (*,G) and (S,G) when static RP configure in
          FHR router
TC_22_2 : Verify OIF and RPF for (*,G) and (S,G) when static RP configure in
          FHR router
TC_23 : Verify (*,G) and (S,G) populated correctly when RPT and SPT path are
        different
TC_24 : Verify (*,G) and (S,G) populated correctly when SPT and RPT share the
        same path
TC_25 : Verify (*,G) and (S,G) populated correctly after clearing the PIM ,
        IGMP and mroutes joins
TC_26 : Restart the PIMd process and verify PIM joins , and mroutes entries
TC_27 : Configure multiple  groups (10 grps) with same RP address
TC_28 : Configure multiple  groups (10 grps) with different RP address
TC_29 : Verify IIF and OIL in updated in mroute when upstream interface
        configure as RP
TC_30 : Verify IIF and OIL change to other path after shut the primary path
TC_31 : Verify RP info and (*,G) mroute after deleting the RP and shut / no
        shut the RPF interface.
TC_32 : Verify RP info and (*,G) mroute after deleting the RP and shut / no
        shut the RPF interface
"""

import os
import sys
import time
import datetime
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# Required to instantiate the topology builder class.

# pylint: disable=C0413
# Import topogen and topotest helpers

from lib.topogen import Topogen, get_topogen
from lib.topolog import logger
from lib.topojson import build_topo_from_json, build_config_from_json

from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    step,
    shutdown_bringup_interface,
    create_static_routes,
)
from lib.pim import (
    create_pim_config,
    verify_igmp_groups,
    verify_upstream_iif,
    verify_join_state_and_timer,
    verify_mroutes,
    verify_pim_neighbors,
    get_pim_interface_traffic,
    verify_pim_rp_info,
    verify_pim_state,
    clear_pim_interface_traffic,
    clear_mroute,
    McastTesterHelper,
)

pytestmark = [pytest.mark.pimd, pytest.mark.staticd]


# Global variables
GROUP_RANGE_ALL = "224.0.0.0/4"
GROUP_RANGE = "225.1.1.1/32"
GROUP_RANGE_LIST_1 = [
    "225.1.1.1/32",
    "225.1.1.2/32",
    "225.1.1.3/32",
    "225.1.1.4/32",
    "225.1.1.5/32",
]
GROUP_RANGE_LIST_2 = [
    "225.1.1.6/32",
    "225.1.1.7/32",
    "225.1.1.8/32",
    "225.1.1.9/32",
    "225.1.1.10/32",
]
GROUP_ADDRESS = "225.1.1.1"
GROUP_ADDRESS_LIST_1 = ["225.1.1.1", "225.1.1.2", "225.1.1.3", "225.1.1.4", "225.1.1.5"]
GROUP_ADDRESS_LIST_2 = [
    "225.1.1.6",
    "225.1.1.7",
    "225.1.1.8",
    "225.1.1.9",
    "225.1.1.10",
]
STAR = "*"
SOURCE_ADDRESS = "10.0.6.2"
SOURCE = "Static"


def build_topo(tgen):
    """Build function"""

    # Building topology from json file
    build_topo_from_json(tgen, TOPO)


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: %s", testsuite_run_time)
    logger.info("=" * 40)

    topology = """

                 _______r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5
                |             |
                |_____________|
                        r4

    """
    logger.info("Master Topology: \n %s", topology)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/multicast_pim_static_rp.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global TOPO
    TOPO = tgen.json_topo

    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    build_config_from_json(tgen, TOPO)

    # Verify PIM neighbors
    result = verify_pim_neighbors(tgen, TOPO)
    assert result is True, "setup_module :Failed \n Error:" " {}".format(result)

    # XXX Replace this using "with McastTesterHelper()... " in each test if possible.
    global app_helper
    app_helper = McastTesterHelper(tgen)

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    app_helper.cleanup()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info("Testsuite end time: %s", time.asctime(time.localtime(time.time())))
    logger.info("=" * 40)


#####################################################
#
#   Testcases
#
#####################################################


def verify_mroute_repopulated(uptime_before, uptime_after):
    """
    API to compare uptime for mroutes

    Parameters
    ----------
    * `uptime_before` : Uptime dictionary for any particular instance
    * `uptime_after` : Uptime dictionary for any particular instance
    """

    for group in uptime_before.keys():
        for source in uptime_before[group].keys():
            if set(uptime_before[group]) != set(uptime_after[group]):
                errormsg = (
                    "mroute (%s, %s) has not come"
                    " up after mroute clear [FAILED!!]" % (source, group)
                )
                return errormsg

            d_1 = datetime.datetime.strptime(uptime_before[group][source], "%H:%M:%S")
            d_2 = datetime.datetime.strptime(uptime_after[group][source], "%H:%M:%S")
            if d_2 >= d_1:
                errormsg = "mroute (%s, %s) is not " "repopulated [FAILED!!]" % (
                    source,
                    group,
                )
                return errormsg

            logger.info("mroute (%s, %s) is " "repopulated [PASSED!!]", source, group)

    return True


def verify_state_incremented(state_before, state_after):
    """
    API to compare interface traffic state incrementing

    Parameters
    ----------
    * `state_before` : State dictionary for any particular instance
    * `state_after` : State dictionary for any particular instance
    """

    for router, state_data in state_before.items():
        for state, _ in state_data.items():
            if state_before[router][state] >= state_after[router][state]:
                errormsg = (
                    "[DUT: %s]: state %s value has not"
                    " incremented, Initial value: %s, "
                    "Current value: %s [FAILED!!]"
                    % (
                        router,
                        state,
                        state_before[router][state],
                        state_after[router][state],
                    )
                )
                return errormsg

            logger.info(
                "[DUT: %s]: State %s value is "
                "incremented, Initial value: %s, Current value: %s"
                " [PASSED!!]",
                router,
                state,
                state_before[router][state],
                state_after[router][state],
            )

    return True


def test_add_delete_static_RP_p0(request):
    """
    TC_1_P0 : Verify upstream interfaces(IIF) and join state are updated
              properly after adding and deleting the static RP
    TC_2_P0 : Verify IIF and OIL in "show ip pim state" updated properly
              after adding and deleting the static RP
    TC_3_P0: (*, G) Mroute entry are cleared when static RP gets deleted
    TC_4_P0: Verify (*,G) prune is send towards the RP after deleting the
             static RP

    Topology used:
         r0------r1-----r2
       iperf    DUT     RP
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("pre-configuration to send IGMP join and multicast traffic")

    step("Enable IGMP on r1 interface and send IGMP " "join (225.1.1.1) to r1")
    step("Configure r2 loopback interface as RP")
    step("Enable PIM between r1 and r3")

    step("r1: Verify show ip igmp group without any IGMP join")
    dut = "r1"
    interface = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, interface, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: IGMP groups should not be present without any IGMP join\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify show ip pim interface traffic without any IGMP join")
    state_dict = {"r1": {"r1-r2-eth1": ["pruneTx"]}}

    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase {} : Failed \n state_before is not dictionary\n Error: {}".format(
        tc_name, result
    )

    step("r0 : Send IGMP join")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify IGMP groups")
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify RP info")
    dut = "r1"
    iif = "r1-r2-eth1"
    rp_address = "1.0.2.17"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)
    step("r1: Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify ip pim join")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    step("r1: Delete RP configuration")

    # Delete RP configuration
    input_dict = {
        "r1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": "1.0.2.17",
                        "group_addr_range": GROUP_RANGE_ALL,
                        "delete": True,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify RP info")
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: RP info should not be present \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream IIF interface {} should not be present\n "
        "Found: {}".format(tc_name, dut, iif, result)
    )

    step("r1: Verify upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream Join State timer should not run\n "
        "Found: {}".format(tc_name, dut, result)
    )

    # 20
    step("r1: Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: PIM state should not be up \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: mroute (*, G) should not be present \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify show ip pim interface traffic without any IGMP join")
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase {} : Failed \n state_before is not dictionary \n Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase{} : Failed Error: {}".format(tc_name, result)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


def test_SPT_RPT_path_same_p1(request):
    """
    TC_24_P1 : Verify (*,G) and (S,G) populated correctly when SPT and RPT
               share the same path

    Topology used:
                ________r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1             r3-----r5

    r1 : LHR
    r2 : RP
    r3 : FHR
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    clear_pim_interface_traffic(tgen, TOPO)

    dut = "r1"
    intf = "r1-r3-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r1-r4-eth3"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "r3"
    intf = "r3-r1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r3-r4-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("Enable IGMP on r1 interface and send IGMP join (225.1.1.1) to R1")
    step("Configure RP on r2 (loopback interface) for the group range" " 224.0.0.0/4")
    step("Enable the PIM on all the interfaces of r1, r2, r3 and r4 routers")
    step("Send multicast traffic from R3")

    step("r2: Verify RP info")
    dut = "r2"
    rp_address = "1.0.2.17"
    iif = "lo"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r0: Send IGMP join")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify IGMP groups")
    dut = "r1"
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r5: Send multicast traffic for group 225.1.1.1")
    result = app_helper.run_traffic("r5", GROUP_ADDRESS, "r3")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = "r1-r2-eth1"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = "r1-r2-eth1"
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (*, G) upstream IIF interface")
    dut = "r2"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes")
    oif = "r2-r1-eth0"
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (S, G) upstream IIF interface")
    iif = "r2-r3-eth1"
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r2: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = "r3-r5-eth3"
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream Join State timer should not run\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = "r3-r2-eth1"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


def test_not_reachable_static_RP_p0(request):
    """
    TC_5_P0: Verify OIF entry for RP is cleared when RP becomes unreachable
    TC_6_P0: Verify IIF and OIL in "show ip pim state" updated properly when
             RP becomes unreachable
    TC_7_P0 : Verify upstream interfaces(IIF) and join state are updated
              properly after adding and deleting the static RP
    TC_8_P0: Verify (*,G) prune is send towards the RP when RP becomes
             unreachable

    Topology used:
         r0------r1-----r2
       iperf    DUT     RP
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    clear_pim_interface_traffic(tgen, TOPO)

    dut = "r1"
    intf = "r1-r3-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "r1"
    intf = "r1-r4-eth3"
    shutdown_bringup_interface(tgen, dut, intf, False)

    step(
        "r1: (*,G) prune is not sent towards the RP interface, verify using"
        "show ip pim interface traffic"
    )
    state_dict = {"r1": {"r1-r2-eth1": ["pruneTx"]}}
    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, state_before
    )

    step("Enable IGMP on r1 interface and send IGMP " "join (225.1.1.1) to r1")
    step("Configure r2 loopback interface as RP")
    step("Enable PIM between r1 and r2")

    step("r0 : Send IGMP join")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify rp info")
    dut = "r1"
    iif = "r1-r2-eth1"
    rp_address = "1.0.2.17"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify IGMP groups")
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 :Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Make RP un-reachable")
    dut = "r1"
    input_dict = {
        dut: {
            "static_routes": [
                {"network": "1.0.2.17/32", "next_hop": "10.0.1.2", "delete": True}
            ]
        }
    }

    result = create_static_routes(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("r1: Check RP detail using show ip pim rp-info OIF should be unknown")
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, "Unknown", rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "r1 : OIL should be same and IIF should be cleared on R1 verify"
        "using show ip pim state"
    )
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: OIL should be same and IIF should be cleared\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: upstream IIF should be unknown , verify using show ip pim" "upstream")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream IIF interface {} should be unknown \n "
        "Found: {}".format(tc_name, dut, iif, result)
    )

    step(
        "r1: join state should not be joined and join timer should stop,"
        "verify using show ip pim upstream"
    )
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream Join State timer should not run\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step(
        "r1: (*,G) prune is sent towards the RP interface, verify using"
        "show ip pim interface traffic"
    )
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase{} : Failed Error: {}".format(tc_name, result)

    step("r1: (*, G) cleared from mroute table using show ip mroute")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: mroute (*, G) should be cleared from mroute table\n "
        "Found: {}".format(tc_name, dut, result)
    )

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


def test_add_RP_after_join_received_p1(request):
    """
    TC_9_P1 : Verify RP configured after IGMP join received, PIM join towards
              RP is sent immediately

    Topology used:
         r0------r1-----r2
       iperf    DUT     RP
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    clear_pim_interface_traffic(tgen, TOPO)

    step("Enable IGMP on R1 interface")
    step("Configure r2 loopback interface as RP")
    step("Enable PIM between r1 and r2")
    step("Delete RP configuration from r1")

    step("r1: Delete RP configuration")
    input_dict = {
        "r1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": "1.0.2.17",
                        "group_addr_range": GROUP_RANGE_ALL,
                        "delete": True,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify rp-info")
    dut = "r1"
    rp_address = "1.0.2.17"
    iif = "r1-r2-eth1"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: RP-info should not be present \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("joinTx value before join sent")
    state_dict = {"r1": {"r1-r2-eth1": ["joinTx"]}}
    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    step("r0 : Send IGMP join (225.1.1.1) to r1, when rp is not configured" "in r1")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: IGMP group is received on R1 verify using show ip igmp groups")
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream IIF interface {} should not be present \n "
        "Found: {}".format(tc_name, dut, iif, result)
    )

    step("r1: Verify upstream join state and join timer")

    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream Join State timer should not run\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: PIM state should not be up\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: mroute (*, G) should not be present in mroute table \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Configure static RP")
    input_dict = {
        "r1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": "1.0.2.17",
                        "group_addr_range": GROUP_RANGE_ALL,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify rp-info")
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)
    logger.info("Expected behavior: %s", result)

    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


def test_reachable_static_RP_after_join_p0(request):
    """
    TC_10_P0 : Verify RP becomes reachable after IGMP join received, PIM join
               towards RP is sent immediately

    Topology used:
         r0------r1-----r3
       iperf    DUT     RP
    """
    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    clear_pim_interface_traffic(tgen, TOPO)

    step("Enable IGMP on r1 interface and send IGMP " "join (225.1.1.1) to r1")
    step("Configure r2 loopback interface as RP")
    step("Enable PIM between r1 and r2")

    step("r1 : Verify pim interface traffic")
    state_dict = {"r1": {"r1-r2-eth1": ["joinTx"]}}
    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, state_before
    )

    step("r1: Make RP un-reachable")
    dut = "r1"
    intf = "r1-r2-eth1"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r1-r3-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r1-r4-eth3"
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1: Verify rp-info")
    rp_address = "1.0.2.17"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_ADDRESS, "Unknown", rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Send IGMP join for 225.1.1.1")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify IGMP groups")
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify upstream IIF interface")
    iif = "r1-r2-eth1"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream IIF interface {} should not be present \n "
        "Found: {}".format(tc_name, dut, iif, result)
    )

    step("r1 : Verify upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream Join State timer should not run\n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1 : Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: PIM state should not be up \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1 : Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: mroute (*, G) should not be present \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step("r1: Make RP reachable")
    intf = "r1-r2-eth1"
    shutdown_bringup_interface(tgen, dut, intf, True)
    intf = "r1-r3-eth2"
    shutdown_bringup_interface(tgen, dut, intf, True)
    intf = "r1-r4-eth3"
    shutdown_bringup_interface(tgen, dut, intf, True)

    step("r1 : Verify rp-info")
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1: Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify ip mroutes")
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)
    logger.info("Expected behavior: %s", result)

    step("r1 : Verify pim interface traffic")
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase{} : Failed Error: {}".format(tc_name, result)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


def test_send_join_on_higher_preffered_rp_p1(request):
    """
     TC_11_P1 : Verify PIM join send towards the higher preferred RP
     TC_12_P1 : Verify PIM prune send towards the lower preferred RP
     TC_13_P1 : Verify RPF interface is updated in mroute (kernel) when higher
                preferred overlapping RP configured
     TC_14_P1 : Verify IIF and OIL in "show ip pim state" updated properly when
                higher preferred overlapping RP configured
     TC_15_P1 : Verify upstream interfaces(IIF) and join state are updated when
                higher preferred overlapping RP is configured
     TC_16_P1 : Verify join is send to lower preferred RP, when higher
                preferred RP gets deleted
     TC_17_P1 : Verify prune is send to higher preferred RP when higher
                preferred RP gets deleted
     TC_18_P1 : Verify RPF interface updated in mroute when higher preferred RP
                gets deleted
     TC_19_P1 : Verify IIF and OIL in "show ip pim state" updated when higher
                preferred overlapping RP is deleted
     TC_20_P1 : Verify PIM upstream IIF updated when higher preferred
                overlapping RP deleted

    Topology used:
                  _______r2
                 |
       iperf     |
         r0-----r1
                 |
                 |_______r4
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    clear_pim_interface_traffic(tgen, TOPO)

    step("Enable IGMP on r1 interface")
    step("Configure RP on r2 (loopback interface) for the group range " "224.0.0.0/4")
    step("Configure RP on r4 (loopback interface) for the group range " "225.1.1.1/32")

    step("r3 : Make all interface not reachable")
    dut = "r3"
    intf = "r3-r1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r3-r2-eth1"
    shutdown_bringup_interface(tgen, dut, intf, False)
    intf = "r3-r4-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "r2"
    intf = "r2-r3-eth1"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "r4"
    intf = "r4-r3-eth1"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "r1"
    intf = "r1-r3-eth2"
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1 : Verify joinTx count before sending join")
    state_dict = {"r1": {"r1-r4-eth3": ["joinTx"], "r1-r2-eth1": ["pruneTx"]}}

    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, state_before
    )

    step("r0 : Send IGMP join for 225.1.1.1")
    result = app_helper.run_join("r0", GROUP_ADDRESS, "r1")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify IGMP groups")
    dut = "r1"
    oif = "r1-r0-eth0"
    result = verify_igmp_groups(tgen, dut, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("Configure static RP for group 225.1.1.1/32")
    input_dict = {
        "r4": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": "1.0.4.17",
                        "group_addr_range": ["225.1.1.1/32"],
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify RP info for group 224.0.0.0/4")
    rp_address_1 = "1.0.2.17"
    iif = "r1-r2-eth1"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address_1, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify RP info for group 225.1.1.1")
    rp_address_2 = "1.0.4.17"
    iif = "r1-r4-eth3"
    result = verify_pim_rp_info(tgen, TOPO, dut, GROUP_RANGE, iif, rp_address_2, SOURCE)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify join is sent to higher preferred RP")
    step("r1 : Verify prune is sent to lower preferred RP")
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase{} : Failed Error: {}".format(tc_name, result)

    step("r1 : Verify ip mroutes")
    iif = "r1-r4-eth3"
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify PIM state")
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify upstream join state and join timer")
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    clear_pim_interface_traffic(tgen, TOPO)

    step("r1 : Verify joinTx, pruneTx count before RP gets deleted")
    state_dict = {"r1": {"r1-r2-eth1": ["joinTx"], "r1-r4-eth3": ["pruneTx"]}}

    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    step("r1 : Delete RP configuration for 225.1.1.1")
    input_dict = {
        "r1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": "1.0.4.17",
                        "group_addr_range": ["225.1.1.1/32"],
                        "delete": True,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify rp-info for group 224.0.0.0/4")
    iif = "r1-r2-eth1"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_ALL, iif, rp_address_1, SOURCE
    )
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("r1 : Verify rp-info for group 225.1.1.1")
    iif = "r1-r4-eth3"
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE, oif, rp_address_2, SOURCE, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: RP-info should not be present \n "
        "Found: {}".format(tc_name, dut, result)
    )

    step(
        "r1 : Verify RPF interface updated in mroute when higher preferred"
        "RP gets deleted"
    )
    iif = "r1-r2-eth1"
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS, iif, oif)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)
    logger.info("Expected behavior: %s", result)

    step(
        "r1 : Verify IIF and OIL in show ip pim state updated when higher"
        "preferred overlapping RP is deleted"
    )
    result = verify_pim_state(tgen, dut, iif, oif, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "r1 : Verify upstream IIF updated when higher preferred overlapping"
        "RP deleted"
    )
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "r1 : Verify upstream join state and join timer updated when higher"
        "preferred overlapping RP deleted"
    )
    result = verify_join_state_and_timer(tgen, dut, iif, STAR, GROUP_ADDRESS)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step(
        "r1 : Verify join is sent to lower preferred RP, when higher"
        "preferred RP gets deleted"
    )
    step(
        "r1 : Verify prune is sent to higher preferred RP when higher"
        " preferred RP gets deleted"
    )
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase{} : Failed \n state_before is not dictionary \n " "Error: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase{} : Failed Error: {}".format(tc_name, result)

    # Uncomment next line for debugging
    # tgen.mininet_cli()

    write_test_footer(tc_name)


if __name__ == "__main__":
    ARGS = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(ARGS))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: NONE
# 2017 by David Lamparter, placed in public domain

import sys, re, subprocess, os

# find all DEFUNs
defun_re = re.compile(
    r"^((DEF(UN(|_ATTR|_CMD_(ELEMENT|FUNC_(DECL|TEXT))|_DEPRECATED|_NOSH|_HIDDEN|SH(|_ATTR|_DEPRECATED|_HIDDEN))?|PY|PY_ATTR|PY_HIDDEN)|ALIAS)\s*\(.*?)^(?=\s*\{)",
    re.M | re.S,
)
define_re = re.compile(r"((^#\s*define[^\n]+[^\\]\n)+)", re.M | re.S)
# find clang-format control that we just inserted
clean_re = re.compile(
    r"^.*/\* \$FRR indent\$ \*/\s*\n\s*/\* clang-format (on|off) \*/\s*\n", re.M
)


def wrap_file(fn):
    with open(fn, "r") as fd:
        text = fd.read()

        repl = (
            r"/* $FRR indent$ */\n/* clang-format off */\n"
            + r"\1"
            + r"/* $FRR indent$ */\n/* clang-format on */\n"
        )

        # around each DEFUN, insert an indent-on/off comment
        text = defun_re.sub(repl, text)
        text = define_re.sub(repl, text)

        ci = subprocess.Popen(
            ["clang-format"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        stdout, ign = ci.communicate(text.encode("utf-8"))
        ci.wait()
        if ci.returncode != 0:
            raise IOError("clang-format returned %d" % (ci.returncode))

        # remove the bits we inserted above
        final = clean_re.sub("", stdout.decode("utf-8"))

        tmpname = fn + ".indent"
        with open(tmpname, "w") as ofd:
            ofd.write(final)
        os.rename(tmpname, fn)


if __name__ == "__main__":
    for fn in sys.argv[1:]:
        wrap_file(fn)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_as_override():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r3 = tgen.gears["r3"]
    r4 = tgen.gears["r4"]

    def _bgp_converge():
        output = json.loads(r3.vtysh_cmd("show ip bgp neighbor 192.168.2.1 json"))
        expected = {
            "192.168.2.1": {
                "bgpState": "Established",
                "addressFamilyInfo": {"ipv4Unicast": {"acceptedPrefixCounter": 2}},
            }
        }
        return topotest.json_cmp(output, expected)

    def _bgp_as_override():
        output = json.loads(r4.vtysh_cmd("show bgp ipv4 unicast json"))
        expected = {
            "routes": {
                "172.16.255.1/32": [{"valid": True, "path": "65003 65002 65003"}]
            }
        }
        return topotest.json_cmp(output, expected)

    step("Initial BGP converge")
    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP convergence on R4"

    step("Check if BGP as-override from R3 works")
    test_func = functools.partial(_bgp_as_override)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see overriden ASN (65001) from R3"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: ISC

# Copyright 2023 6WIND S.A.
# Authored by Farid Mihoub <farid.mihoub@6wind.com>
#

# IANA Address Family Identifier
AFI_IP = 1
AFI_IP6 = 2
AFI_L2VPN = 25

# IANA Subsequent Address Family Idenitifier
SAFI_UNICAST = 1
SAFI_MULTICAST = 2
SAFI_MPLS_LABEL = 4
SAFI_EVPN = 70
SAFI_MPLS_VPN = 128
SAFI_IP_FLOWSPEC = 133
SAFI_VPN_FLOWSPEC = 134


# ------------------------------------------------------------------------------
class AddressFamily:
    def __init__(self, afi, safi):
        self.afi = afi
        self.safi = safi

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.afi, self.safi) == (other.afi, other.safi)

    def __str__(self):
        return f"afi: {self.afi}, safi: {self.safi}"

    def __hash__(self):
        return hash((self.afi, self.safi))


# ------------------------------------------------------------------------------
class AF:
    IPv4_UNICAST = AddressFamily(AFI_IP, SAFI_UNICAST)
    IPv6_UNICAST = AddressFamily(AFI_IP6, SAFI_UNICAST)
    IPv4_VPN = AddressFamily(AFI_IP, SAFI_MPLS_VPN)
    IPv6_VPN = AddressFamily(AFI_IP6, SAFI_MPLS_VPN)
    IPv4_MPLS = AddressFamily(AFI_IP, SAFI_MPLS_LABEL)
    IPv6_MPLS = AddressFamily(AFI_IP6, SAFI_MPLS_LABEL)
    IPv4_FLOWSPEC = AddressFamily(AFI_IP, SAFI_IP_FLOWSPEC)
    IPv6_FLOWSPEC = AddressFamily(AFI_IP6, SAFI_IP_FLOWSPEC)
    VPNv4_FLOWSPEC = AddressFamily(AFI_IP, SAFI_VPN_FLOWSPEC)
    VPNv6_FLOWSPEC = AddressFamily(AFI_IP6, SAFI_VPN_FLOWSPEC)
    L2EVPN = AddressFamily(AFI_L2VPN, SAFI_EVPN)
    L2VPN_FLOWSPEC = AddressFamily(AFI_L2VPN, SAFI_VPN_FLOWSPEC)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_pbr_topo1.py
#
# Copyright (c) 2020 by
# Cumulus Networks, Inc.
# Donald Sharp
#
# Copyright (c) 2023 LabN Consulting, L.L.C.
#

"""
test_pbr_topo1.py: Testing PBR

"""

import os
import sys
import pytest
import json
import platform
import re
from functools import partial

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger
from lib.common_config import shutdown_bringup_interface

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.pbrd]

#####################################################
##
##   Network Topology Definition
##
#####################################################


def build_topo(tgen):
    "Build function"

    # Populate routers
    for routern in range(1, 2):
        tgen.add_router("r{}".format(routern))

    # Populate switches
    for switchn in range(1, 6):
        switch = tgen.add_switch("sw{}".format(switchn))
        switch.add_link(tgen.gears["r1"])


#####################################################
##
##   Tests starting
##
#####################################################


def setup_module(module):
    "Setup topology"
    tgen = Topogen(build_topo, module.__name__)
    tgen.start_topology()

    krel = platform.release()
    if topotest.version_cmp(krel, "4.10") < 0:
        tgen.errors = "Newer kernel than 4.9 needed for pbr tests"
        pytest.skip(tgen.errors)

    router_list = tgen.routers()
    for rname, router in router_list.items():
        # Install vrf into the kernel and slave eth3
        router.run("ip link add vrf-chiyoda type vrf table 1000")
        router.run("ip link set dev {}-eth3 master vrf-chiyoda".format(rname))
        router.run("ip link set vrf-chiyoda up")

        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_PBRD, os.path.join(CWD, "{}/pbrd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def test_converge_protocols():
    "Wait for protocol convergence"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    topotest.sleep(5, "Waiting for PBR convergence")


#
# router: r1
# tag: "show pbr interface"
# cmd: "show pbr interface json"
# expfile: "{}/{}/pbr-interface.json".format(CWD, router.name)
#
def runit(router, tag, cmd, expfile):
    logger.info(expfile)

    # Read expected result from file
    expected = json.loads(open(expfile).read())

    # Actual output from router
    test_func = partial(topotest.router_json_cmp, router, cmd, expected)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assertmsg = '"{}" mismatches on {}'.format(tag, router.name)
    if result is not None:
        gather_pbr_data_on_error(router)
        assert result is None, assertmsg


def test_pbr_data():
    "Test PBR"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Verify PBR Status
    logger.info("Verifying PBR routes")

    router_list = tgen.routers().values()
    for router in router_list:
        runit(
            router,
            "show pbr interface",
            "show pbr interface json",
            "{}/{}/pbr-interface.json".format(CWD, router.name),
        )

        runit(
            router,
            "show pbr map",
            "show pbr map json",
            "{}/{}/pbr-map.json".format(CWD, router.name),
        )

        runit(
            router,
            "show pbr nexthop-groups",
            "show pbr nexthop-groups json",
            "{}/{}/pbr-nexthop-groups.json".format(CWD, router.name),
        )


########################################################################
# 			Field test - START
########################################################################

#
# New fields:
# match ip-protocol (was only tcp|udp, now any value in /etc/protocols)
# match pcp (0-7)
# match vlan (1-4094)
# match vlan (tagged|untagged|untagged-or-zero)
#

#
# c:   command
# cDN: omit default destination IP address (special case)
# tm:  must-match pattern
# tN:  must Not match pattern
#
# Note we are searching amid a bunch of other rules, so these elements
# should be unique.
#
ftest = [
    {"c": "match ip-protocol icmp", "tm": r"IP protocol Match: 1$"},
    {"c": "no match ip-protocol icmp", "tN": r"IP protocol Match:"},
    {"c": "match pcp 6", "tm": r"PCP Match: 6$"},
    {"c": "match pcp 0", "tm": r"PCP Match: 0$"},
    {"c": "no match pcp 0", "tN": r"PCP Match:"},
    {"c": "match vlan 33", "tm": r"VLAN ID Match: 33$"},
    {"c": "no match vlan 33", "tN": r"VLAN ID Match:"},
    {"c": "match vlan tagged", "tm": r"VLAN Flags Match: tagged$"},
    {"c": "match vlan untagged", "tm": r"VLAN Flags Match: untagged$"},
    {"c": "match vlan untagged-or-zero", "tm": r"VLAN Flags Match: untagged-or-zero$"},
    {"c": "no match vlan tagged", "tN": r"VLAN Flags Match:"},
    {"c": "match src-ip 37.49.22.0/24", "tm": r"SRC IP Match: 37.49.22.0/24$"},
    {"c": "no match src-ip 37.49.22.0/24", "tN": r"SRC IP Match: 37.49.22.0/24$"},
    {
        "c": "match dst-ip 38.41.29.0/25",
        "cDN": "foo",
        "tm": r"DST IP Match: 38.41.29.0/25$",
    },
    {"c": "no match dst-ip 38.41.29.0/25", "tN": r"DST IP Match: 38.41.29.0/25$"},
    {"c": "match src-port 117", "tm": r"SRC Port Match: 117$"},
    {"c": "no match src-port 117", "tN": r"SRC Port Match: 117$"},
    {"c": "match dst-port 119", "tm": r"DST Port Match: 119$"},
    {"c": "no match dst-port 119", "tN": r"DST Port Match: 119$"},
    {"c": "match dscp cs3", "tm": r"DSCP Match: 24$"},
    {"c": "no match dscp cs3", "tN": r"DSCP Match: 24$"},
    {"c": "match dscp 5", "tm": r"DSCP Match: 5$"},
    {"c": "no match dscp 5", "tN": r"DSCP Match: 5$"},
    {"c": "match ecn 2", "tm": r"ECN Match: 2$"},
    {"c": "no match ecn 2", "tN": r"ECN Match: 2$"},
    {"c": "match mark 337", "tm": r"MARK Match: 337$"},
    {"c": "no match mark 337", "tN": r"MARK Match: 337$"},
    {"c": "set src-ip 44.100.1.1", "tm": r"Set SRC IP: 44.100.1.1$"},
    {"c": "no set src-ip 44.100.1.1", "tN": r"Set SRC IP: 44.100.1.1$"},
    {"c": "set dst-ip 44.105.1.1", "tm": r"Set DST IP: 44.105.1.1$"},
    {"c": "no set dst-ip 44.105.1.1", "tN": r"Set DST IP: 44.105.1.1$"},
    {"c": "set src-port 41", "tm": r"Set SRC PORT: 41$"},
    {"c": "no set src-port 41", "tN": r"Set SRC PORT: 41$"},
    {"c": "set dst-port 43", "tm": r"Set DST PORT: 43$"},
    {"c": "no set dst-port 43", "tN": r"Set DST PORT: 43$"},
    {"c": "set dscp 24", "tm": r"Set DSCP: 24$"},
    {"c": "no set dscp 24", "tN": r"Set DSCP: 24$"},
    {"c": "set dscp cs7", "tm": r"Set DSCP: 56$"},
    {"c": "no set dscp cs7", "tN": r"Set DSCP: 56$"},
    {"c": "set ecn 1", "tm": r"Set ECN: 1$"},
    {"c": "no set ecn 1", "tN": r"Set ECN: 1$"},
]


# returns None if command output is correct, otherwise returns output
def rtr_field_cmp(rtr, cmd, pat_mustmatch, pat_mustnotmatch):
    outstr = rtr.vtysh_cmd(cmd)
    if pat_mustmatch is not None:
        logger.info("MUSTMATCH: {}".format(pat_mustmatch))
        m = re.search(pat_mustmatch, outstr, flags=re.M)
        if not m:
            logger.info('Missing MUSTMATCH "{}"'.format(pat_mustmatch))
            return "MISSING MUSTMATCH: " + outstr
    if pat_mustnotmatch is not None:
        logger.info("MUSTNOTMATCH: {}".format(pat_mustnotmatch))
        m = re.search(pat_mustnotmatch, outstr, flags=re.M)
        if m:
            logger.info('Has MUSTNOTMATCH "{}"'.format(pat_mustnotmatch))
            return "HAS MUSTNOTMATCH: " + outstr
    return None


#
# This test sets fields in pbrd and looks for them in zebra via "sh pbr map"
#
def test_pbr_fields():
    "Test setting and clearing rule fields"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Verifying PBR rule fields")

    # uncomment for manual interaction
    # tgen.cli()

    tag = "field"

    router_list = tgen.routers().values()
    for router in router_list:
        for t in ftest:
            # send field-setting command
            # always have a match dst-ip to satisfy rule non-empty check
            if "cDN" in t:
                match_dstip = ""
            else:
                match_dstip = "match dst-ip 9.9.9.9/32\n"
            vcmd = "c t\npbr-map ASAKUSA seq 100\n{}\n{}set nexthop-group A\nend\nend".format(
                t["c"], match_dstip
            )
            router.vtysh_multicmd(vcmd)

            # debug
            router.vtysh_cmd("sh pbr map")

            match = None
            notmatch = None

            if "tm" in t:
                match = t["tm"]
                logger.info("MUSTMATCH: {}".format(match))
            if "tN" in t:
                notmatch = t["tN"]
                logger.info("NOTMATCH: {}".format(notmatch))

            test_func = partial(rtr_field_cmp, router, "sh pbr rule", match, notmatch)
            _, result = topotest.run_and_expect(test_func, None, count=10, wait=1)
            assertmsg = '"{}" mismatches on {}'.format(tag, router.name)
            if result is not None:
                gather_pbr_data_on_error(router)
                assert result is None, assertmsg

        #
        # clean up
        #
        vcmd = "c t\nno pbr-map ASAKUSA seq 100\nend"
        router.vtysh_multicmd(vcmd)

        match = None
        notmatch = r"Seq 100\w"

        test_func = partial(rtr_field_cmp, router, "sh pbr rule", match, notmatch)
        _, result = topotest.run_and_expect(test_func, None, count=10, wait=1)
        assertmsg = '"{}" mismatches on {}'.format(tag, router.name)
        if result is not None:
            gather_pbr_data_on_error(router)
            assert result is None, assertmsg


########################################################################
# 			Field test - END
########################################################################


def test_pbr_flap():
    "Test PBR interface flapping"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Verify PBR Status
    logger.info("Flapping PBR Interfaces")

    router_list = tgen.routers().values()
    for router in router_list:
        # Flap interface to see if route-map properties are intact
        # Shutdown interface

        for i in range(5):
            intf = "r1-eth{}".format(i)

            # Down and back again
            shutdown_bringup_interface(tgen, router.name, intf, False)
            shutdown_bringup_interface(tgen, router.name, intf, True)

        intf_file = "{}/{}/pbr-interface.json".format(CWD, router.name)
        logger.info(intf_file)

        # Read expected result from file
        expected = json.loads(open(intf_file).read())

        # Actual output from router
        test_func = partial(
            topotest.router_json_cmp, router, "show pbr interface json", expected
        )
        _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = '"show pbr interface" mismatches on {}'.format(router.name)
        if result is not None:
            gather_pbr_data_on_error(router)
            assert result is None, assertmsg


def test_rule_linux_installation():
    "Ensure that rule is installed in the kernel"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking for installed PBR rules in OS")

    def _get_router_rules(router, expected):
        actual = topotest.ip_rules(router)

        logger.info(actual)
        return topotest.json_cmp(actual, expected)

    router_list = tgen.routers().values()
    for router in router_list:
        rules_file = "{}/{}/linux-rules.json".format(CWD, router.name)

        expected = json.loads(open(rules_file).read())

        test_func = partial(_get_router_rules, router, expected)

        _, result = topotest.run_and_expect(test_func, None, count=20, wait=1)
        assertmsg = "Router {} OS rules mismatch".format(router.name)
        assert result is None, assertmsg


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))


#
# EXTRA SAUCE
#
def gather_pbr_data_on_error(router):
    logger.info(router.vtysh_cmd("show ip route"))
    logger.info(router.vtysh_cmd("show ip route vrf vrf-chiyoda"))
    logger.info(router.vtysh_cmd("show ip nht"))
    logger.info(router.vtysh_cmd("show pbr interface"))
    logger.info(router.vtysh_cmd("show pbr map"))
    logger.info(router.vtysh_cmd("show pbr nexthop-groups"))
    logger.info(router.vtysh_cmd("show nexthop-group rib singleton ip"))
    logger.info(router.vtysh_cmd("show nexthop-group rib singleton ipv6"))
    logger.info(router.vtysh_cmd("show nexthop-group rib"))
    logger.info(router.run("ip nexthop show"))
    logger.info(router.run("ip route show"))
    logger.info(router.run("ip route show table 1000"))
    logger.info(router.run("ip route show table 10000"))
    logger.info(router.run("ip -6 route show table 10000"))
    logger.info(router.run("ip route show table 10001"))
    logger.info(router.run("ip -6 route show table 10001"))
    logger.info(router.run("ip route show table 10002"))
    logger.info(router.run("ip -6 route show table 10002"))
    logger.info(router.run("ip route show table 10003"))
    logger.info(router.run("ip -6 route show table 10003"))
    logger.info(router.run("ip route show table 10004"))
    logger.info(router.run("ip -6 route show table 10004"))
    logger.info(router.run("ip route show table 10005"))
    logger.info(router.run("ip -6 route show table 10005"))
    logger.info(router.run("ip rule show"))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2021 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#

"""
Test if we send ONLY GUA address for route-server-client peers.
"""

import os
import sys
import json
import pytest
from functools import partial
import functools

pytestmark = [pytest.mark.bgpd]

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen


def build_topo(tgen):
    """
    All peers are FRR BGP peers except r5 that is a exabgp peer.
    Exabgp does not send any IPv6 Link-Local nexthop

    r2 is a route-server view RS AS 65000
    Other routers rX has AS 6500X

                     +---+
                     | r3|
                     +---+
                       |
                2001:db8:3::0/64
                       |
                     eth1
                     +---+
                     |r2 |
                     +---+
                     eth0
                       |
                 2001:db8:1::0/64
                  /    |     \
              +---+  +---+  +---+
              | r1|  | r4|  |r5 |
              +---+  +---+  +---+
    """

    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    sw1 = tgen.add_switch("s1")
    sw1.add_link(tgen.gears["r1"])
    sw1.add_link(tgen.gears["r2"])
    sw1.add_link(tgen.gears["r4"])

    sw2 = tgen.add_switch("s2")
    sw2.add_link(tgen.gears["r2"])
    sw2.add_link(tgen.gears["r3"])

    ## Add iBGP ExaBGP neighbor
    peer_ip = "2001:db8:1::4"  ## peer
    peer_route = "via 2001:db8:1::1"  ## router
    r5 = tgen.add_exabgp_peer("r5", ip=peer_ip, defaultRoute=peer_route)
    sw1.add_link(r5)


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()

    # Start r5 exabgp peer
    r5 = tgen.gears["r5"]
    r5.start(os.path.join(CWD, "r5"), os.path.join(CWD, "exabgp.env"))


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def get_link_local(rname, ifname, cache):
    ip = cache.get(rname, {}).get(ifname)
    if ip:
        return ip

    tgen = get_topogen()
    out = tgen.gears[rname].vtysh_cmd("show interface %s json" % ifname, isjson=True)
    for address in out[ifname]["ipAddresses"]:
        if not address["address"].startswith("fe80::"):
            continue
        ip = address["address"].split("/")[0]
        cache.setdefault(rname, {})[ifname] = ip
        return ip


def replace_link_local(expected, cache):
    for prefix, prefix_infos in expected.get("routes", {}).items():
        for prefix_info in prefix_infos:
            for nexthop in prefix_info.get("nexthops", []):
                ip = nexthop.get("ip", "")
                if not ip.startswith("link-local:"):
                    continue
                rname = ip.split(":")[1]
                ifname = ip.split(":")[2]
                ip = get_link_local(rname, ifname, cache)
                nexthop["ip"] = ip


def check_r2_sub_group(expected):
    tgen = get_topogen()

    r2 = tgen.gears["r2"]

    output = json.loads(r2.vtysh_cmd("show bgp view RS update-groups json"))
    actual = [
        subgroup["peers"]
        for entry in output.get("RS", {}).values()
        for subgroup in entry["subGroup"]
    ]

    return topotest.json_cmp(actual, expected)


def test_converge_protocols():
    "Wait for protocol convergence"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    ref_file = "{}/{}/show_bgp_ipv6_summary.json".format(CWD, r2.name)
    expected = json.loads(open(ref_file).read())

    test_func = partial(
        topotest.router_json_cmp,
        r2,
        "show bgp view RS ipv6 summary json",
        expected,
    )
    _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assertmsg = "{}: BGP convergence failed".format(r2.name)
    assert res is None, assertmsg


def test_bgp_route_server_client_step1():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    global link_local_cache
    link_local_cache = {}
    router_list = tgen.routers().values()
    for router in router_list:
        if router.name == "r2":
            # route-server
            cmd = "show bgp view RS ipv6 unicast json"
        else:
            cmd = "show bgp ipv6 unicast json"

        ref_file = "{}/{}/show_bgp_ipv6_step1.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            cmd,
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 table failure".format(router.name)
        assert res is None, assertmsg

    # check r2 sub-groups
    expected = [["2001:db8:1::4"], ["2001:db8:1::3", "2001:db8:1::2", "2001:db8:3::2"]]

    test_func = functools.partial(check_r2_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


def test_bgp_route_server_client_step2():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    r2.vtysh_cmd(
        """
configure terminal
router bgp 65000 view RS
 address-family ipv6 unicast
  neighbor 2001:db8:1::2 nexthop-local unchanged
  neighbor 2001:db8:1::3 nexthop-local unchanged
  neighbor 2001:db8:1::4 nexthop-local unchanged
  neighbor 2001:db8:3::2 nexthop-local unchanged
"""
    )

    router_list = tgen.routers().values()
    for router in router_list:
        if router.name == "r2":
            # route-server
            cmd = "show bgp view RS ipv6 unicast json"
        else:
            cmd = "show bgp ipv6 unicast json"

        ref_file = "{}/{}/show_bgp_ipv6_step2.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            cmd,
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 table failure".format(router.name)
        assert res is None, assertmsg

    # check r2 sub-groups
    expected = [
        ["2001:db8:1::4"],
        ["2001:db8:1::3", "2001:db8:1::2"],
        ["2001:db8:3::2"],
    ]

    test_func = functools.partial(check_r2_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


def test_bgp_route_server_client_step3():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    r2.vtysh_cmd(
        """
configure terminal
router bgp 65000 view RS
 address-family ipv6 unicast
  no neighbor 2001:db8:1::2 nexthop-local unchanged
  no neighbor 2001:db8:1::3 nexthop-local unchanged
  no neighbor 2001:db8:1::4 nexthop-local unchanged
  no neighbor 2001:db8:3::2 nexthop-local unchanged
"""
    )

    global link_local_cache
    link_local_cache = {}
    router_list = tgen.routers().values()
    for router in router_list:
        if router.name == "r2":
            # route-server
            cmd = "show bgp view RS ipv6 unicast json"
        else:
            cmd = "show bgp ipv6 unicast json"

        ref_file = "{}/{}/show_bgp_ipv6_step1.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            cmd,
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 table failure".format(router.name)
        assert res is None, assertmsg

    # check r2 sub-groups
    expected = [["2001:db8:1::4"], ["2001:db8:1::3", "2001:db8:1::2", "2001:db8:3::2"]]

    test_func = functools.partial(check_r2_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_bgp_nexthop_ipv6_topo1.py
#
# Copyright (c) 2024 by
# Cumulus Networks, Inc.
# 6WIND S.A.
#

"""
Ensure that BGP ipv6 nexthops are correct
"""

import os
import sys
import pytest
from functools import partial
import json

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.


pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    r"""
    All peers are FRR BGP peers except r3 that is a exabgp peer.
    rr is a route-reflector for AS 65000 iBGP peers.
    Exabgp does not send any IPv6 Link-Local nexthop

                   iBGP peers        |  eBGP peers
                                     |
                    AS 65000         |
                                     |
                     +---+           |
                     | r6|           |
                     +---+           |
                       |             |
             fd00:0:3::0/64          |
                       |             |  AS 65002
                     +---+           |         +---+
                     |rr |----fd00:0:4::0/64---| r5|
                     +---+           |_________+---+
                       |             |         +---+
                 fd00:0:2::0/64----------------| r4|
                  /    |     \       |         +---+
              +---+  +---+  +---+    |  AS 65001
              | r1|  | r2|  |r3 |    |
              +---+  +---+  +---+
    """

    def connect_routers(tgen, left, right):
        for rname in [left, right]:
            if rname not in tgen.routers().keys():
                tgen.add_router(rname)

        switch = tgen.add_switch("s-{}-{}".format(left, right))
        switch.add_link(tgen.gears[left], nodeif="eth-{}".format(right))
        switch.add_link(tgen.gears[right], nodeif="eth-{}".format(left))

    def connect_switchs(tgen, rname, switch):
        if rname not in tgen.routers().keys():
            tgen.add_router(rname)

        switch.add_link(tgen.gears[rname], nodeif="eth-{}".format(switch.name))

    def connect_dummy(tgen, rname, switch):
        if rname not in tgen.routers().keys():
            tgen.add_router(rname)

        switch.add_link(tgen.gears[rname], nodeif="eth-dummy")

    # sw_du switch is for a dummy interface (for local network)
    for i in range(1, 7):
        if i == 3:
            # r3 is an exabgp peer
            continue
        sw_du = tgen.add_switch("sw%s" % i)
        connect_dummy(tgen, "r%s" % i, sw_du)

    # sw switch is for interconnecting peers on the same subnet
    sw = tgen.add_switch("sw")
    connect_switchs(tgen, "rr", sw)
    connect_switchs(tgen, "r1", sw)
    connect_switchs(tgen, "r2", sw)
    connect_switchs(tgen, "r4", sw)

    # directly connected without switch routers
    connect_routers(tgen, "rr", "r5")
    connect_routers(tgen, "rr", "r6")

    ## Add iBGP ExaBGP neighbor
    peer_ip = "fd00:0:2::3"  ## peer
    peer_route = "via fd00:0:2::9"  ## router
    r3 = tgen.add_exabgp_peer("r3", ip=peer_ip, defaultRoute=peer_route)
    sw.add_link(r3)


#####################################################
##
##   Tests starting
##
#####################################################


def setup_module(module):
    "Setup topology"
    tgen = Topogen(build_topo, module.__name__)
    tgen.start_topology()

    # This is a sample of configuration loading.
    router_list = tgen.routers()
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()

    # Start r3 exabgp peer
    r3 = tgen.gears["r3"]
    r3.start(os.path.join(CWD, "r3"), os.path.join(CWD, "exabgp.env"))


def get_link_local(rname, ifname, cache):
    ip = cache.get(rname, {}).get(ifname)
    if ip:
        return ip

    tgen = get_topogen()
    out = tgen.gears[rname].vtysh_cmd("show interface %s json" % ifname, isjson=True)
    for address in out[ifname]["ipAddresses"]:
        if not address["address"].startswith("fe80::"):
            continue
        ip = address["address"].split("/")[0]
        cache.setdefault(rname, {})[ifname] = ip
        return ip


def replace_link_local(expected, cache):
    for prefix, prefix_info in expected.get("routes", {}).items():
        for nexthop in prefix_info[0].get("nexthops", []):
            ip = nexthop.get("ip", "")
            if not ip.startswith("link-local:"):
                continue
            rname = ip.split(":")[1]
            ifname = ip.split(":")[2]
            ip = get_link_local(rname, ifname, cache)
            nexthop["ip"] = ip


def check_rr_sub_group(expected):
    tgen = get_topogen()

    rr = tgen.gears["rr"]

    output = json.loads(rr.vtysh_cmd("show bgp update-groups json"))
    actual = [
        subgroup["peers"]
        for entry in output.get("default", {}).values()
        for subgroup in entry["subGroup"]
    ]

    return topotest.json_cmp(actual, expected)


def teardown_module(_mod):
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def test_converge_protocols():
    "Wait for protocol convergence"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rr = tgen.gears["rr"]
    ref_file = "{}/{}/show_bgp_ipv6_summary.json".format(CWD, rr.name)
    expected = json.loads(open(ref_file).read())

    test_func = partial(
        topotest.router_json_cmp,
        rr,
        "show bgp ipv6 summary json",
        expected,
    )
    _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assertmsg = "{}: BGP convergence".format(rr.name)
    assert res is None, assertmsg


def test_bgp_ipv6_table_step1():
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    global link_local_cache
    link_local_cache = {}
    router_list = tgen.routers().values()
    for router in router_list:
        ref_file = "{}/{}/show_bgp_ipv6_step1.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show bgp ipv6 unicast json",
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 Nexthop failure".format(router.name)
        assert res is None, assertmsg

    # check rr sub-groups
    expected = [
        ["fd00:0:2::1", "fd00:0:2::2"],
        ["fd00:0:2::3"],
        ["fd00:0:2::4"],
        ["fd00:0:3::5"],
        ["fd00:0:4::6"],
    ]

    test_func = partial(check_rr_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


def test_bgp_ipv6_table_step2():
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rr = tgen.gears["rr"]
    rr.vtysh_cmd(
        """
configure terminal
router bgp 65000
 address-family ipv6 unicast
  no neighbor fd00:0:2::1 nexthop-local unchanged
  no neighbor fd00:0:2::2 nexthop-local unchanged
  no neighbor fd00:0:2::3 nexthop-local unchanged
  no neighbor fd00:0:2::4 nexthop-local unchanged
  no neighbor fd00:0:3::5 nexthop-local unchanged
  no neighbor fd00:0:4::6 nexthop-local unchanged
"""
    )

    router_list = tgen.routers().values()
    for router in router_list:
        ref_file = "{}/{}/show_bgp_ipv6_step2.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show bgp ipv6 unicast json",
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 Nexthop failure".format(router.name)
        assert res is None, assertmsg

    # check rr sub-groups
    expected = [
        ["fd00:0:2::1", "fd00:0:2::2"],
        ["fd00:0:2::3"],
        ["fd00:0:3::5", "fd00:0:2::4"],
        ["fd00:0:4::6"],
    ]

    test_func = partial(check_rr_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


def test_bgp_ipv6_table_step3():
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rr = tgen.gears["rr"]
    rr.vtysh_cmd(
        """
configure terminal
router bgp 65000
 address-family ipv6 unicast
  neighbor fd00:0:2::1 nexthop-local unchanged
  neighbor fd00:0:2::2 nexthop-local unchanged
  neighbor fd00:0:2::3 nexthop-local unchanged
  neighbor fd00:0:2::4 nexthop-local unchanged
  neighbor fd00:0:3::5 nexthop-local unchanged
  neighbor fd00:0:4::6 nexthop-local unchanged
"""
    )

    router_list = tgen.routers().values()
    for router in router_list:
        ref_file = "{}/{}/show_bgp_ipv6_step1.json".format(CWD, router.name)
        expected = json.loads(open(ref_file).read())
        replace_link_local(expected, link_local_cache)

        test_func = partial(
            topotest.router_json_cmp,
            router,
            "show bgp ipv6 unicast json",
            expected,
        )
        _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assertmsg = "{}: BGP IPv6 Nexthop failure".format(router.name)
        assert res is None, assertmsg

    # check rr sub-groups
    expected = [
        ["fd00:0:2::1", "fd00:0:2::2"],
        ["fd00:0:2::3"],
        ["fd00:0:2::4"],
        ["fd00:0:3::5"],
        ["fd00:0:4::6"],
    ]

    test_func = partial(check_rr_sub_group, expected)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Peer group split failed"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright 2023 6WIND S.A.
# Authored by Farid Mihoub <farid.mihoub@6wind.com>
#

"""
test_bgp_bmp.py: Test BGP BMP functionalities

    +------+            +------+               +------+
    |      |            |      |               |      |
    | BMP1 |------------|  R1  |---------------|  R2  |
    |      |            |      |               |      |
    +------+            +------+               +------+

Setup two routers R1 and R2 with one link configured with IPv4 and
IPv6 addresses.
Configure BGP in R1 and R2 to exchange prefixes from
the latter to the first router.
Setup a link between R1 and the BMP server, activate the BMP feature in R1
and ensure the monitored BGP sessions logs are well present on the BMP server.
"""

from functools import partial
import json
import os
import pytest
import sys

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join("../"))
sys.path.append(os.path.join("../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.bgp import verify_bgp_convergence_from_running_config
from lib.bgp import bgp_configure_prefixes
from .bgpbmp import (
    bmp_check_for_prefixes,
    bmp_check_for_peer_message,
    bmp_update_seq,
)
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

pytestmark = [pytest.mark.bgpd]

PRE_POLICY = "pre-policy"
POST_POLICY = "post-policy"
LOC_RIB = "loc-rib"

UPDATE_EXPECTED_JSON = False
DEBUG_PCAP = False


def build_topo(tgen):
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_bmp_server("bmp1", ip="192.0.2.10", defaultRoute="via 192.0.2.1")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["bmp1"])

    tgen.add_link(tgen.gears["r1"], tgen.gears["r2"], "r1-eth1", "r2-eth0")


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    if DEBUG_PCAP:
        pcap_file = os.path.join(tgen.logdir, "r1/bmp.pcap")
        tgen.gears["r1"].run(
            "tcpdump -nni r1-eth0 -s 0 -w {} &".format(pcap_file), stdout=None
        )

    for _, (rname, router) in enumerate(tgen.routers().items(), 1):
        logger.info("Loading router %s" % rname)
        router.load_frr_config(
            os.path.join(CWD, "{}/frr.conf".format(rname)),
            [(TopoRouter.RD_ZEBRA, None), (TopoRouter.RD_BGP, "-M bmp")],
        )

    tgen.start_router()

    logger.info("starting BMP servers")
    for bmp_name, server in tgen.get_bmp_servers().items():
        server.start(log_file=os.path.join(tgen.logdir, bmp_name, "bmp.log"))


def teardown_module(_mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_convergence():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    result = verify_bgp_convergence_from_running_config(tgen, dut="r1")
    assert result is True, "BGP is not converging"


def _test_prefixes(policy, vrf=None, step=0):
    """
    Setup the BMP  monitor policy, Add and withdraw ipv4/v6 prefixes.
    Check if the previous actions are logged in the BMP server with the right
    message type and the right policy.
    """
    tgen = get_topogen()

    safi = "vpn" if vrf else "unicast"

    prefixes = ["172.31.0.15/32", "2001::1111/128"]

    for type in ("update", "withdraw"):
        bmp_update_seq(tgen.gears["bmp1"], os.path.join(tgen.logdir, "bmp1", "bmp.log"))

        bgp_configure_prefixes(
            tgen.gears["r2"],
            65502,
            "unicast",
            prefixes,
            vrf=vrf,
            update=(type == "update"),
        )

        logger.info(f"checking for prefixes {type}")

        for ipver in [4, 6]:
            if UPDATE_EXPECTED_JSON:
                continue
            ref_file = "{}/r1/show-bgp-ipv{}-{}-step{}.json".format(
                CWD, ipver, type, step
            )
            expected = json.loads(open(ref_file).read())

            test_func = partial(
                topotest.router_json_cmp,
                tgen.gears["r1"],
                f"show bgp ipv{ipver} {safi} json",
                expected,
            )
            _, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
            assertmsg = f"r1: BGP IPv{ipver} convergence failed"
            assert res is None, assertmsg

        # check
        test_func = partial(
            bmp_check_for_prefixes,
            prefixes,
            type,
            policy,
            step,
            tgen.gears["bmp1"],
            os.path.join(tgen.logdir, "bmp1"),
            tgen.gears["r1"],
            f"{CWD}/bmp1",
            UPDATE_EXPECTED_JSON,
            LOC_RIB,
        )
        success, res = topotest.run_and_expect(test_func, None, count=30, wait=1)
        assert success, "Checking the updated prefixes has failed ! %s" % res


def test_bmp_server_logging():
    """
    Assert the logging of the bmp server.
    """

    def check_for_log_file():
        tgen = get_topogen()
        output = tgen.gears["bmp1"].run(
            "ls {}".format(os.path.join(tgen.logdir, "bmp1"))
        )
        if "bmp.log" not in output:
            return False
        return True

    success, _ = topotest.run_and_expect(check_for_log_file, True, count=30, wait=1)
    assert success, "The BMP server is not logging"


def test_peer_up():
    """
    Checking for BMP peers up messages
    """

    tgen = get_topogen()
    peers = ["192.168.0.2", "192:168::2", "0.0.0.0"]

    logger.info("checking for BMP peers up messages")

    test_func = partial(
        bmp_check_for_peer_message,
        peers,
        "peer up",
        tgen.gears["bmp1"],
        os.path.join(tgen.logdir, "bmp1", "bmp.log"),
    )
    success, _ = topotest.run_and_expect(test_func, True, count=30, wait=1)
    assert success, "Checking the updated prefixes has been failed !."


def test_bmp_bgp_unicast():
    """
    Add/withdraw bgp unicast prefixes and check the bmp logs.
    """
    logger.info("*** Unicast prefixes pre-policy logging ***")
    _test_prefixes(PRE_POLICY, step=1)
    logger.info("*** Unicast prefixes post-policy logging ***")
    _test_prefixes(POST_POLICY, step=1)
    logger.info("*** Unicast prefixes loc-rib logging ***")
    _test_prefixes(LOC_RIB, step=1)


def test_bmp_bgp_vpn():
    # check for the prefixes in the BMP server logging file
    logger.info("***** VPN prefixes pre-policy logging *****")
    _test_prefixes(PRE_POLICY, vrf="vrf1", step=2)
    logger.info("***** VPN prefixes post-policy logging *****")
    _test_prefixes(POST_POLICY, vrf="vrf1", step=2)
    logger.info("***** VPN prefixes loc-rib logging *****")
    _test_prefixes(LOC_RIB, vrf="vrf1", step=2)


def test_peer_down():
    """
    Checking for BMP peers down messages
    """
    tgen = get_topogen()

    tgen.gears["r2"].vtysh_cmd("clear bgp *")

    peers = ["192.168.0.2", "192:168::2"]

    logger.info("checking for BMP peers down messages")

    test_func = partial(
        bmp_check_for_peer_message,
        peers,
        "peer down",
        tgen.gears["bmp1"],
        os.path.join(tgen.logdir, "bmp1", "bmp.log"),
    )
    success, _ = topotest.run_and_expect(test_func, True, count=30, wait=1)
    assert success, "Checking the updated prefixes has been failed !."


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
